/*
	Driver for the Towitoko Chipdrive Micro 120.
	
	This file is part of Smartk, a smartcard framework for the Linux Kernel.
	Copyright (C) 2003	Luigi Catuogno		<luicat@dia.unisa.it>
						Roberto Gassira'	<robgas@dia.unisa.it>
						Michele Masullo		<micmas@dia.unisa.it>
	
    Smartk is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Smartk is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "smartk.h"
#include <linux/mm.h>			/*to the end of includes necessary!*/

/* Definizione macro per licenza */
#define MY_AUTHOR       "Roberto Gassira' <robgas@tcfs.it>"
#define MY_DESC         "Modulo per ifd towitoko"

#define IFD_TOWITOKO_LED_OFF            0x00
#define IFD_TOWITOKO_LED_RED            0x01
#define IFD_TOWITOKO_CLOCK_RATE         (372L * 9600L)
#define IFD_TOWITOKO_MAX_TRANSMIT       255


//#define ATR_MAX_SIZE    				33  /* Maximum size of ATR byte array */


/* Identificazione del codice e della sua licenza*/
MODULE_AUTHOR(MY_AUTHOR);
MODULE_DESCRIPTION(MY_DESC);
MODULE_LICENSE("GPL");

/*	Puntatore alla memoria contenente il valore dell'atr della carta	*/
static char * icc_atr;

/*	Puntatore al core */
static struct smartk *mycore;
	
/*	Dichiarazione funzioni interne al modulo	*/
//static int ifd_smartk_set_color_LED(int color);
static BYTE ifd_smartk_check_command(BYTE * command, unsigned int size);
static int ifd_smartk_sendrec(BYTE *send, int lens, BYTE *receive, int lenr);
static int ifd_smartk_activate_card();
static int ifd_smartk_deactivate_card();
static int ifd_smartk_set_bitrate(int bitrate);
static int ifd_smartk_set_color_LED(int color);
static int ifd_smartk_check_reader();
static int ifd_smartk_card_present();
static int ifd_smartk_reset_card(char * retatr, int *len);
static void ifd_smartk_blinking(int ntime, long timeout);

/*	Dichiarazione funzioni registrate dal modulo	*/
static int ifd_smartk_init_reader(struct smartk *core);
static int ifd_smartk_cleanup_reader(struct smartk *core);	
static int ifd_smartk_set_parity(struct smartk *core, int parity);
static int ifd_smartk_activate_icc(struct smartk *core, int with_ATR, char **res_ATR, int *len);
static int ifd_smartk_deactivate_icc(struct smartk *core);
static int ifd_smartk_transmit(struct smartk *core, BYTE* buffer, int count);
static int ifd_smartk_receive(struct smartk *core, BYTE* buffer, int count);
	
/*	Registrazione struttura ifd	*/
static struct ifd_smartk towitoko = {
	.init_reader = ifd_smartk_init_reader,
	.cleanup_reader = ifd_smartk_cleanup_reader,
	.activate_card = ifd_smartk_activate_icc,
	.deactivate_card = ifd_smartk_deactivate_icc,
	.set_parity = ifd_smartk_set_parity,
	.transmit = ifd_smartk_transmit,
	.receive = ifd_smartk_receive
};

/*	Calcolo del ckecksum del comando	*/
static BYTE ifd_smartk_check_command(BYTE * command, unsigned int size){
	BYTE checksum, x7;
	int i;
	
	checksum = 0;
	for (i = 0; i < size; i++){
    	checksum = checksum ^ command[i];
      	x7 = (checksum & 0x80) >> 7;
      	checksum = checksum << 1;
      	checksum = !x7 == 0x01 ? checksum | 0x01 : checksum & 0xFE;
    }

  return checksum;
}

/*	Read/Write con la seriale	*/
static int ifd_smartk_sendrec(BYTE *send, int lens, BYTE *receive, int lenr){
	
	/*	Inviamo il comando*/
	if ( mycore->io->write(mycore, send, lens, 0) != OK){
		PERROR("ifd_towitoko:set_bitrate->Errore nella write del comando\n");
		return FAIL;		
	}
	
	/*	Lettura della risposta con timeout in millisecondi	*/
	if ( mycore->io->read(mycore, receive, lenr, 1000) != OK){
		PERROR("ifd_towitoko:set_bitrate->Errore nella read della risposta comando\n");
		return FAIL;
	}
	
	return OK;
}

/*	Attivazione della carta	*/
static int ifd_smartk_activate_card(){
	BYTE status[1];
	BYTE buffer[3] = { 0x60, 0x0F, 0x9C };
	
	/*	Calcolo del checksum del comando	*/
	buffer[2] = ifd_smartk_check_command(buffer, 2);	
		
	if	(ifd_smartk_sendrec(buffer, 3, status, 1) != OK){
		PERROR("ifd_towitoko:activate_card->Errore nel comando\n");
		return FAIL;
	}

	/*	Controllo della risposta	*/
	if ( status[0] != 0x01){
		PERROR("ifd_towitoko:activate_card->Errore risposta setting status=%02X\n", status[0]);
		return FAIL;
	}
	
	return OK;
}

/*	Disattivazione della carta	*/
static int ifd_smartk_deactivate_card(){
	BYTE status[1];
	BYTE buffer[3] = { 0x61, 0x0F, 0x98 };
	
	/*	Calcolo del checksum del comando	*/
	buffer[2] = ifd_smartk_check_command(buffer, 2);	
		
	if	(ifd_smartk_sendrec(buffer, 3, status, 1) != OK){
		PERROR("ifd_towitoko:deactivate_card->Errore nel comando\n");
		return FAIL;
	}

	/*	Controllo della risposta	*/
	if ( status[0] != 0x01){
		PERROR("ifd_towitoko:deactivate_card->Errore risposta setting status=%02X\n", status[0]);
		return FAIL;
	}
	
	return OK;
}


/*	Funzioni esportate dal modulo	*/
/*	Per il futuro ma tanto non funziona sul programmatore 	*/
static int ifd_smartk_set_bitrate(int bitrate){
	BYTE status[1];
	BYTE buffer[6] = { 0x6E, 0x00, 0x00, 0x00, 0x08, 0x00};
	/*Setting velocita' per 9600*/
	buffer[1] = 0x0B;
    buffer[3] = 0x02;
	
	/* Magia */
	buffer[2] = buffer[1] ^ 0x5D;

	/*	Calcolo del checksum del comando	*/
	buffer[5] = ifd_smartk_check_command(buffer, 5);
	
	/*	Inviamo il comando*/
	if ( mycore->io->write(mycore, buffer, 6, 0) != OK){
		PERROR("ifd_towitoko:set_bitrate->Errore nella write del comando\n");
		return FAIL;		
	}
	
	/*	Lettura della risposta con timeout in millisecondi	*/
	if ( mycore->io->read(mycore, status, 1, 1000) != OK){
		PERROR("ifd_towitoko:set_bitrate->Errore nella read della risposta comando\n");
		return FAIL;
	}
	
	/*	Controllo della risposta	*/
	if ( status[0] != 0x01){
		PERROR("ifd_towitoko:set_bitrate->Errore risposta setting status=%02X\n", status[0]);
		return FAIL;
	}
	
	return OK;
} 


int ifd_smartk_set_parity(struct smartk *core, int parity){
	BYTE buffer[5] = { 0x6F, 0x00, 0x6A, 0x0F, 0x00 };
  	BYTE status[1];
	
	/*	Set del parity	*/
	if(parity == IO_SERIAL_PARITY_ODD)
		buffer[1] = 0x80;
	else
		buffer[1] = 0x40;
	
	/*	Calcolo del checksum	*/
	buffer[4] = ifd_smartk_check_command(buffer, 4);
	
	/*	Inviamo il comando*/
	if ( mycore->io->write(mycore, buffer, 5, 0) != OK){
		PERROR("ifd_towitoko:set_parity->Errore nella write del comando\n");
		return FAIL;		
	}
	
	/*	Lettura della risposta con timeout in millisecondi	*/
	if ( mycore->io->read( mycore, status, 1, 1000) != OK){
		PERROR("ifd_towitoko:set_parity->Errore nella read della risposta comando\n");
		return FAIL;
	}
	
	/*	Controllo della risposta	*/
	if ( status[0] != 0x01){
		PERROR("ifd_towitoko:set_parity->Errore risposta setting status=%02X\n", status[0]);
		/*	Nuovo tentativo di set parity	*/
		if	(ifd_smartk_sendrec(buffer, 5, status, 1) != OK){
			PERROR("ifd_towitoko:set_parity->Errore comunicazione nel secondo tentativo!\n");
			return FAIL;
		}
		if ( status[0] != 0x01){
			PERROR("ifd_towitoko:set_parity->Errore risposta secondo setting status=%02X\n", status[0]);
			return FAIL;
		}
	}
	
	PDEBUG("ifd_towitoko:set_parity->Parity = %s\n", parity == IO_SERIAL_PARITY_ODD \
			? "ODD" : "EVEN");
	
	/*	Aggiorniamo il valore del parity e attivazione led	*/
	if( towitoko.parity != parity ){
		towitoko.parity = parity;
		ifd_smartk_set_color_LED(IFD_TOWITOKO_LED_RED);
	}
	
	return OK;
}


/*	Controllo sul programmatore connesso	*/
static int ifd_smartk_check_reader(){
	BYTE status[3];
  	BYTE buffer[2] = { 0x00, 0x01};
	
	buffer[1] = ifd_smartk_check_command(buffer, 1);
	
	if	(ifd_smartk_sendrec(buffer, 2, status, 3) != OK){
		PERROR("ifd_towitoko:check_reader->Errore nel comando\n");
		return FAIL;
	}

	if(status[0] == 0x61)
		PDEBUG("ifd_towitoko:check_reader->OK! Reader = %s\n", towitoko.reader);
	return OK;
}

/*	Configurazione Led del programmatore	*/
static int ifd_smartk_set_color_LED(int color){
	BYTE status[1];
  	BYTE buffer[5] = { 0x6F, 0x00, 0x6A, 0x0F, 0x00 };

	buffer[1] = color;

	buffer[4] = ifd_smartk_check_command(buffer, 4);
	
	/*	Inviamo il comando*/
	if ( mycore->io->write(mycore, buffer, 5, 0) != OK){
		PERROR("ifd_towitoko:set_color_LED->Errore nella write del comando\n");
		return FAIL;		
	}
	
	/*	Lettura della risposta con timeout in millisecondi	*/
	if ( mycore->io->read(mycore, status, 1, 1000) != OK){
		PERROR("ifd_towitoko:set_color_LED->Errore nella read della risposta comando\n");
		return FAIL;
	}
	
	/*	Controllo della risposta	*/
	if ( status[0] != 0x01){
		PERROR("ifd_towitoko:set_color_LED->Errore risposta setting status=%02X\n", status[0]);
		return FAIL;
	}
	
	PDEBUG("ifd_towitoko:set_color_LED->Settato led\n");
	return OK;
}


/*	Test per vedere se la carta e' presente	*/
static int ifd_smartk_card_present(){
	BYTE buffer[2] = { 0x03, 0x07 };
  	BYTE status[2];
	
	buffer[1] = ifd_smartk_check_command(buffer, 1);
	
	/*	Inviamo il comando*/
	if ( mycore->io->write(mycore, buffer, 2, 0) != OK){
		PERROR("ifd_towitoko:card_present->Errore nella write del comando\n");
		return FAIL;		
	}
	
	/*	Lettura della risposta con timeout in millisecondi	*/
	if ( mycore->io->read(mycore, status, 2, 1000) != OK){
		PERROR("ifd_towitoko:card_present->Errore nella read del comando\n");
		return FAIL;
	}
	/*	Controllo della risposta: se status[0]==0xC0 0 0x80
		Ripeto l'operazione*/
	if(	(status[0] == 0xc0) || (status[0] == 0x80)){
		if ( mycore->io->write(mycore, buffer, 2, 0) != OK){
			PERROR("ifd_towitoko:card_present->Errore nella write del comando\n");
			return FAIL;		
		}
	
		/*	Lettura della risposta con timeout in millisecondi	*/
		if ( mycore->io->read(mycore, status, 2, 1000) != OK){
			PERROR("ifd_towitoko:card_present->Errore nella read del comando\n");
			return FAIL;
		}
	}
	
	PDEBUG("ifd_towitoko:card_present->Byte[0]=%02X\n", status[0]);
	
	/*	Controllo della risposta	*/
	if ( status[0] == 0x40)
			return OK;
	else 
			return FAIL;
}

/*	Reset della carta	*/
static int ifd_smartk_reset_card(char * retatr, int *len){
	BYTE buffer[5] = { 0xA0, 0x6F, 0x00, 0x05, 0x74 };
	BYTE status[ATR_MAX_SIZE];
	int i, j, res;
	
	buffer[4] = ifd_smartk_check_command(buffer, 4);
	
	if ( mycore->io->write(mycore, buffer, 5, 0) != OK){
		PERROR("ifd_towitoko:reset_card->Errore nella write del comando\n");
		return FAIL;		
	}
	
	/*	Inizio ricezione risposta	*/
	for(i = 0; i < ATR_MAX_SIZE; i++){
		/*	Lettura in timeout, fine ATR	*/
		if( (res = mycore->io->read(mycore, status+i, 1, 1000)) == -ETIME)
			break;
		else if(res == FAIL )
			return FAIL;
						
	}	
	
	/*	Controllo sulla convention della carta	*/
	if(status[0] == 0x03){
		/*	Inversione dei byte della carta */
		for(j = 0; j < i; j++)
			status[j] = ~( INVERT_BYTE (status[j]) );			
	}
	
	/*	Visualizzazione Output	*/
	PDEBUG("ifd_towitoko:reset_card->ATR = %s \n", mycore->xpdu_to_string(status, i));
	
	/*	Generazione ATR	*/
	if( i < 2)
		return FAIL;
	else
		memcpy(retatr, status, i);
	
	/*	Informiamo della lunghezza dell'ATR	*/
	*len = i;
	
	return  OK;
}


/*	Blinking del Led	*/
static void ifd_smartk_blinking(int ntime, long timeout){
	int i=0;
	for(i=0; i< ntime; i++){
		ifd_smartk_set_color_LED(IFD_TOWITOKO_LED_RED);
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout( (timeout/1000)*(HZ));
		ifd_smartk_set_color_LED(IFD_TOWITOKO_LED_OFF);
	}	
}



/*---------------------->Funzioni esterne <-------------------*/
/*	Init del modulo	*/
static int ifd_smartk_init(void){
	
	/*	Carimamento modulo per caricamento della seriale	*/
	/*	DA FARE	*/
	
	/*	Init struttura ifd	*/
	towitoko.clock_rate = IFD_TOWITOKO_CLOCK_RATE;
	strcpy(towitoko.port_mod, "serial");
	strcpy(towitoko.reader, "Towitoko Chipdrive Micro");
	
	/*	Registrazione del device driver per Towitoko	*/
	if( register_ifd_smartk(&towitoko) != OK ){
		PERROR("ifd_towitoko:init->Impossibile registrare Towitoko!\n");
		goto error;		
	}
		
	PINFO("ifd_towitoko:init->IFD Layer On!\n");
	//PDEBUG("ifd_towitoko:init->towitoko=%X\n", (int)&towitoko);
	return OK;
	
error_load_mod:
	PDEBUG("ifd_towitoko:init->Impossile trovare il modulo io_smartk\n");
	
error:
	return(FAIL);
}


/*	Attivazione del lettore	*/
static int ifd_smartk_init_reader(struct smartk *core){
	struct io_property serial_prop;
	mycore = core;
	PDEBUG("ifd_towitoko:init_reader!\n");
	
	
	/*	Attivazione della porta seriale	*/
	if( core->io->init_port(core) != OK){
		PERROR("ifd_towitoko:init_reader-> Impossibile arrivare porta seriale!\n");
		return FAIL;		
	}
	
	serial_prop.bitrate = B9600;
	serial_prop.bits = CS8;
	serial_prop.stopbits = 2;
	serial_prop.parity = towitoko.parity = IO_SERIAL_PARITY_EVEN;
	serial_prop.dtr = IO_SERIAL_HIGH; 
	serial_prop.rts = IO_SERIAL_HIGH;
	
	
	/*	Configurazione porta seriale	*/
	if( core->io->set_property(core, &serial_prop) != OK){
		PERROR("ifd_towitoko:init_reader-> Impossibile configurare porta seriale!\n");
		return FAIL;		
	}
	
	/*	Attivazione parity	*/
	if( ifd_smartk_set_parity(core, serial_prop.parity) != OK){
		PERROR("ifd_towitoko:init->Errore nel Set parity!\n");
		return FAIL;
	}
	
	/*	Controllo del programmatore connesso	*/
	if( ifd_smartk_check_reader() != OK){
		PERROR("ifd_towitoko:Reader non supportato!\n");
		return FAIL;
	}
	
	/*	Segnale di attivazione Reader	*/
	ifd_smartk_blinking(2,800);
	
	PINFO("ifd_towitoko:init_reader-> Towitoko Ready!\n");
	return OK;	
}


/*	Disattivazione del lettore	*/
static int ifd_smartk_cleanup_reader(struct smartk *core){
	/*	Disattivazione Carta	*/
	ifd_smartk_deactivate_card();
	
	/*	Spegniamo led programmatore	*/
	ifd_smartk_blinking(1,1500);
	
	/*	Sblocco della porta	*/
	if( core->io->cleanup_port(core) != OK)
		PERROR("ifd_towitoko:init_reader-> Impossibile sbloccare la porta!\n");	

	/*	Annullamento del core	*/
	mycore = NULL;
	
	PINFO("ifd_towitoko:cleanup_reader-> Towitoko Off!\n");
	
	return OK;	
}


/*	Attivazione smartcard	*/
static int ifd_smartk_activate_icc(struct smartk *core, int with_ATR, char **res_ATR, int *len){
	icc_atr = (char *)kmalloc(ATR_MAX_SIZE, GFP_KERNEL);
	int rlen;
	
	PDEBUG("ifd_towitoko:activate_icc!\n");
	
	/*	Controllo Carta presente	*/
	if( ifd_smartk_card_present() != OK ){
		PERROR("ifd_towitoko:activate_icc->Carta non presente, Inserire Carta!\n");
		return FAIL;		
	}
	
	/*	Attivazione Carta	*/
	if(ifd_smartk_activate_card() != OK){
		PERROR("ifd_towitoko:activate_icc->Impossibile attivare la carta!\n");
		return FAIL;		
	}
	
	/*	Reset della Carta	*/
	if( ifd_smartk_reset_card(icc_atr, &rlen) != OK ){
		PERROR("ifd_towitoko:activate_icc->Impossibile resettare la carta!\n");
		return FAIL;		
	}
	
	/*	Led ON	*/
	ifd_smartk_set_color_LED(IFD_TOWITOKO_LED_RED);
	
	
	/*Preparazione della risposta	*/
	if(with_ATR == TRUE){
		*res_ATR = icc_atr;
		*len = rlen;
	}else
		/*	Pulizia della memoria	*/
		kfree(icc_atr);

	return OK;
}


/*	Disattivazione carta	*/
static int ifd_smartk_deactivate_icc(struct smartk *core){
	/*	Disattiviamo carta*/
	if( ifd_smartk_deactivate_card() != OK ){
		PERROR("ifd_towitoko:deactivate_icc-> Errore nella disattivazione della carta!\n");
		return FAIL;		
	}
	
	/*	Led OFF	*/
	ifd_smartk_set_color_LED(IFD_TOWITOKO_LED_OFF);
	
	return OK;
}

/*	Invio dati	*/
static int ifd_smartk_transmit(struct smartk *core, BYTE* buffer, int count){
	BYTE header[6] = {0x6F, 0x00, 0x05, 0x00, 0xFE, 0xF8};
	int to_send;
	
	/*	Calcolo dati da mandare	*/
	to_send = MIN(count, IFD_TOWITOKO_MAX_TRANSMIT);
	
	/*	Preparazione del lettore a ricevere i dati	*/
	header[1] = (BYTE) to_send;
	header[3] = ifd_smartk_check_command(header, 3);
	
	/*=========Attenzione! forse si devono mandare 4/6 byte da leggere!=======*/
	if ( core->io->write(core, header, 4, 0) != OK){
		PERROR("ifd_towitoko:transmit->Errore nella write del comando\n");
		return FAIL;		
	}
	
	/*	Gestione di piu' dati!	*/
	/*	Invio reale dei dati	*/
	if ( core->io->write(core, buffer, to_send, core->timings.char_delay) != OK){
		PERROR("ifd_towitoko:transmit->Errore nella write del comando\n");
		return FAIL;		
	}
	
	return OK;
}

/*	Ricezione dati	*/
static int ifd_smartk_receive(struct smartk *core, BYTE* buffer, int count){
	/*	Gestione semplificata senza controllo del primo carattere*/
	if( core->io->read( core, buffer, count, core->timings.char_timeout) != OK){
		PERROR("ifd_towitoko:receive->Errore nella read del comando\n");
		return FAIL;
	}
	return OK;
}


static void ifd_smartk_exit(void){
	PINFO("ifd_towitoko:init->IFD Layer Off!\n");
}


module_init(ifd_smartk_init);
module_exit(ifd_smartk_exit);
