/*
	Simple interface for the serial port.
	
	This file is part of Smartk, a smart card framework for the Linux Kernel.
	Copyright (C) 2003	Luigi Catuogno		<luicat@dia.unisa.it>
						Roberto Gassira'	<robgas@dia.unisa.it>
						Michele Masullo		<micmas@dia.unisa.it>
	
    Smartk is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Smartk is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
	
#include "smartk.h"

#include <asm/uaccess.h>  			/* x set_fs ecc */
#include <asm/termios.h>			/* x ioctl */
#include <linux/sched.h>			/*Per conoscere la variabile current*/
#include <linux/fs.h>				/*per file*/
#include <linux/timer.h>			/*fdsfsdfasfsdf*/
#include <linux/mm.h>				/*x Attenzione! Includere per ultimo */

/* Definizione macro per licenza */
#define MY_AUTHOR       "Roberto Gassira' <robgas@tcfs.it>"
#define MY_DESC         "Modulo di base per interfacciare la seriale"

/* Identificazione del codice e della sua licenza*/
MODULE_AUTHOR(MY_AUTHOR);
MODULE_DESCRIPTION(MY_DESC);
MODULE_LICENSE("GPL");

/* Macro interne*/
#define DEF_TTYZ        	"/dev/ttyS0"
#define TTYZ_LEN			11
#define READ_BUF_LEN 		1024           	/* Max length of the message from the device */
#define OPEN_FLAGS			O_RDWR|O_NOCTTY
#define N_TTY_BUF_SIZE		4096

/*  Dichiarazione dei prototipi delle funzioni */
static int io_smartk_init(void);
static void io_smartk_cleanup(void);

static int io_smartk_init_port(struct smartk * core);
static int io_smartk_cleanup_port(struct smartk * core);

static int io_smartk_read(struct smartk * core, BYTE * data, int count, unsigned timeout);
static int io_smartk_write(struct smartk * core, BYTE * data, int count, unsigned timeout);
	
static int io_smartk_get_property(struct smartk * core, struct io_property *newprop);
static int io_smartk_set_property(struct smartk * core, const struct io_property *newprop);
	
/*	Dichiarazione variabili globali interne (ovvero definite static)*/
static struct tty_ldisc io_smartk_ldisc;

/*	Gestione dei dati in arrivo	*/
static struct {
	BYTE data[READ_BUF_LEN];
	int head;
	int tail;
	int nbyte;
	int ask;
	struct semaphore sem;
	spinlock_t lock;
}read_buf;

/*	Gestore della seriale	*/
static struct{
	short active;
	struct file * file;
  	struct file_operations *fops;
  	struct tty_struct *tty;
	struct inode *inode;
}io_serial;

/*	Struttura per registrazione	*/
static struct io_smartk serial = {
	.init_port = io_smartk_init_port,
	.cleanup_port = io_smartk_cleanup_port,
	.read = io_smartk_read,
	.write = io_smartk_write,
	.get_property = io_smartk_get_property,
	.set_property = io_smartk_set_property
};
	
/*Si definisce questa nuova macro quando si configura la velocita' di input della seriale 
  uguale a 0 (la velocita' della seriale e' la stessa per input/output
*/
#define IBAUD0  			020000000000

/* Parametri per il modulo */
static char *inputdev;
MODULE_PARM(inputdev, "s");
MODULE_PARM_DESC(inputdev, "Device da leggere");

/*	Funzioni Interne	*/
/*	Visualizza il bitrate attivato	*/
static char * baud_to_string(int rate){
	char *ret = (char *)kmalloc(10, GFP_KERNEL);
	switch(rate){
		case B9600:		strcpy(ret, "B9600"); break;
		case B1200:		strcpy(ret, "B1200"); break;
		default:	strcpy(ret, "ERROR"); break;
	}
	return ret;
}

/*
 *	Definizione funzioni Line Discipline
 */
/*Apertura del line discipline*/
static int io_smartk_ldisc_open(struct tty_struct *tty){
	PDEBUG("io_serial:ldisc_open-> apertura seriale\n");
	
	/*	Controllo sul tty	*/
	if(io_serial.tty != tty){
		PDEBUG("io_serial:ldisc_open->Strutture tty differenti!\n");
		return(FAIL);	
	}
	
	/*	Controllo se porta attiva	*/
	if(io_serial.active == TRUE)
		return FAIL;		
	
	return OK;
}


/*Chiusura del line discipline*/
static void io_smartk_ldisc_close(struct tty_struct *tty)
{
	PDEBUG("io_serial:ldisc_close-> chiusura seriale\n");
}


/*	Gestione dati in arrivo	*/
static void io_smartk_receive_buf(struct tty_struct *tty, const unsigned char *cp, \
									char *fp, int count){
	unsigned long cpuflags;
	int i;
	char buffer_msg[READ_BUF_LEN];
	
	/*	Controllo su attivazione lettura	*/
	if(io_serial.active){									
		
		/*	Entrata in sezione critica	*/
		down_interruptible(&read_buf.sem);
	
		/*	Blocco degli interrupt per non corrompere i dati nel buffer	*/
		spin_lock_irqsave(&read_buf.lock, cpuflags);
	
		/*	Controllo carta levata	*/
		if( (cp[0] == 0x00) && (count == 1))
			PDEBUG("io_serial:receive_buf->Carta Espulsa!\n");	
		else{
			i = MIN(count, READ_BUF_LEN - read_buf.head);
			memcpy(read_buf.data+read_buf.head, cp, i);
			read_buf.head += i;
			read_buf.nbyte += i;
		}
		/*	Sblocco interrupt	*/
		spin_unlock_irqrestore(&read_buf.lock, cpuflags);
	
		/*	Output dati letti */
		for(i=0; i<count; i++)
			sprintf(buffer_msg+i*3, "%02X ", cp[i]);
		buffer_msg[i*3] = '\0';
		
		PDEBUG("io_serial:receive_buf->New Data:Len=%d Data=%s\n", count, buffer_msg);	
		//PDEBUG("io_serial:receive_buf->read_buf=%s\n", xpdu_to_string(read_buf.data+read_buf.tail, read_buf.nbyte));
		
		
		/*Svegliamo qualcuno che stava dormendo*/
		if(read_buf.ask)
			wake_up_interruptible(&io_serial.tty->read_wait);			
		
		/*	Uscita sezione critica	*/
		up(&read_buf.sem);
	}
	
	
}


/*	Funzione che viene chiamata dal driver per conoscere quanti dati 
*	vengono gestiti dalla receive_buf.
*/
static int io_smartk_receive_room(struct tty_struct  *tty){
	
	return READ_BUF_LEN;

}


/*	Funzioni per il modulo	*/

/*	Caricamento del modulo per registrare la nuova line discipline 
*	e per attivarla sulla seriale io_smartk_ttyz */	
static int io_smartk_init()
{
	int ret;
	char *io_smartk_ttyz = (char *)kmalloc(TTYZ_LEN, GFP_KERNEL);
	mm_segment_t oldfs;
	int ldisc_number = N_MOUSE;
	
	/* Definizione della nuova line discipline*/
	memset(&io_smartk_ldisc, 0, sizeof(io_smartk_ldisc));
    io_smartk_ldisc.magic  = TTY_LDISC_MAGIC;
    io_smartk_ldisc.name   = "io_smartk";
    io_smartk_ldisc.flags  = 0;
    io_smartk_ldisc.open   = io_smartk_ldisc_open;
    io_smartk_ldisc.close  = io_smartk_ldisc_close;
    io_smartk_ldisc.read   = NULL;
    io_smartk_ldisc.write  = NULL;
    io_smartk_ldisc.ioctl  = NULL;
    io_smartk_ldisc.poll   = NULL;
    io_smartk_ldisc.receive_buf = io_smartk_receive_buf;
    io_smartk_ldisc.receive_room = io_smartk_receive_room;
    io_smartk_ldisc.write_wakeup = NULL;
	
	/*Registrazione*/
	if (( ret = tty_register_ldisc(N_MOUSE, &io_smartk_ldisc)) != 0 ){
		PDEBUG("io_serial:init->Impossibile registrare ldisc! Errore=%d\n", ret);
		return ret;
	}
	
	/* Recupero il puntatore a file_operations di /dev/ttyS0 */
	if (inputdev)
		strncpy(serial.device, inputdev, MIN(MAXIODEVNAME, strlen(inputdev)));
	else
		strncpy(serial.device, DEF_TTYZ, TTYZ_LEN);
				
	io_serial.file = filp_open(serial.device, OPEN_FLAGS, 0700);
	
	if (IS_ERR(io_serial.file)){
     	const int open_errno = -PTR_ERR(io_serial.file);
     	PDEBUG("io_serial:init->Impossibile aprire %s:errore=%d\n", io_smartk_ttyz, open_errno);
		return open_errno;
    }
	
	io_serial.fops = io_serial.file->f_op;
	io_serial.active = FALSE;
	
	/*	Attivazione del nuovo line discipline della seriale	*/
	io_serial.tty = (struct tty_struct *)io_serial.file->private_data;
	io_serial.inode = io_serial.file->f_dentry->d_inode;
	
	oldfs = get_fs();
    set_fs(get_ds());
	ret = io_serial.fops->ioctl(io_serial.inode, io_serial.file, TIOCSETD, (unsigned long)&ldisc_number);
	if (ret < 0){
		PDEBUG("io_serial:init->Errore %d su device %s\n", ret, serial.device);	
		return FAIL;
	}
	set_fs(oldfs);
	
	/*	Inizializzazione della coda di attesa	*/
	init_waitqueue_head(&io_serial.tty->read_wait);
	
	/*	Inizializzazione buffer lettura*/
	read_buf.head = read_buf.tail = read_buf.nbyte = read_buf.ask = 0;
	sema_init(&read_buf.sem, 1);
	spin_lock_init(&read_buf.lock);
	
	/*	Registrazione porta comunicazione per smartk	*/
	if(register_io_smartk(&serial) != OK){
		PDEBUG("io_serial:init->Impossibile registrare porta seriale!\n");
		io_smartk_cleanup();
		return FAIL;
	}
	
	PINFO("io_serial:init->IO Layer(%s) ON!\n", serial.device);
	
	return OK;
}


/*	Blocco seriale	*/
static int io_smartk_init_port(struct smartk * core){
	
	io_serial.active = TRUE;
	
	PDEBUG("io_serial:init_port-> Porta Attivata\n");
 	return OK;	
}

/*	Sblocco seriale	*/
static int io_smartk_cleanup_port(struct smartk * core){
	
	io_serial.active = FALSE;
	
	PDEBUG("io_serial:cleanup_port-> Porta Disattivata\n");
 	return OK;	
}

/* 	Lettura della seriale ricordandoci di come funzionano le regole
 *	del line discipline. */
static int io_smartk_read(struct smartk * core, BYTE * data, int count, unsigned timeout){
	/*	Riportiamo il valore del timeout in millisecondi*/
	int res, jms = (timeout/1000) * HZ;
	
	/*	Entrata sezione critica	*/
	down_interruptible( &read_buf.sem);
	
	//PDEBUG("read_buf.nbyte = %d, count = %d, read_buf.head = %d \n", read_buf.nbyte, count, read_buf.head);

	/*	Controllo dati presenti nel buffer	*/
	while (read_buf.nbyte < count){	
		
		/*	Indichiamo la richiesta di lettura	*/
		read_buf.ask = 1;
	
		/*	Uscita dalla sezione critica per aggiornare buffer	*/
		up(&read_buf.sem);
		
		//PDEBUG("io_smaerk:read->Aspetto arrivo dati\n");
		/*	Blocco per ricezione dati	*/
		if( (res = interruptible_sleep_on_timeout(&io_serial.tty->read_wait, jms)) < 0){
			PDEBUG("io_smaerk:read->Uscita coda per errore\n");
			return -ERESTARTSYS;
		}		
	
		if (!res)
			PDEBUG("io_serial:read->Uscira coda di attesa per timeout!\n");
		else
			PDEBUG("io_serial:read->Uscira coda di attesa per wakeup dopo %d ms!\n", (jms - res));	
	
		/*	Lettura andata in timeout	*/
		if( (count > 0) && (res == 0)){
			PDEBUG("io_smaerk:read->Dati non ARRIVATI?\n");
			return -ETIME;			
		}
		
		/*	Rientriamo in sezione critica	*/
		down_interruptible( &read_buf.sem);
		
		
		/*	Controllo per verificare se si sono ricevuti meno dati di quanti richiesti	*/
		if (read_buf.nbyte < count){
			PDEBUG("io_serial:read->Sono arrivati meno byte!\n");
		}
		
	}
	
	/*	Fine richiesta di read	*/
	read_buf.ask = 0;
	
	/*	Copia della risposta e aggiornamento valori	*/
	memcpy(data, read_buf.data+read_buf.tail, count);
	read_buf.nbyte -= count;
	
	if (read_buf.head == (read_buf.tail + count))
		read_buf.head = read_buf.tail = 0;
	else
		read_buf.tail += count;
	
	/*	Uscita sezione critica*/
	up(&read_buf.sem);
	
	return OK;
}


/* 	Scrittura sulla seriale utilizzando direttamente il driver	*/
static int io_smartk_write(struct smartk * core, BYTE * data, int count, unsigned timeout){
	int ret;
	
	//PDEBUG("io_serial:write-> Richiesta scrittura %s\n", core->xpdu_to_string(data, count));
	ret = io_serial.tty->driver.write(io_serial.tty, 0, data, count );
	/*Controllo sulla risposta*/
	if ( (ret <= 0) || (ret != count)){
		PDEBUG("io_serial:write-> errore write - ret = %d count = %d\n", ret, count);
		return FAIL;
	}
	
	return OK;
}

/*	Prendiamo tutte le informazioni della porta al fine di garantire
*	la comunicazione diretta con il programmatore
*/
static int io_smartk_get_property(struct smartk * core, struct io_property *newprop){
	struct termios * serial_termios = io_serial.tty->termios;
		
	PDEBUG("io_serial:get_serial_prop->Line Discipline = %d\n", io_serial.tty->termios->c_line);
	
	/*	Controlli opzionali	*/
	/*	Controllo speed	*/
	if( serial.property.bitrate != C_BAUD(io_serial.tty) )
		PDEBUG("io_serial:get_serial_prop->Baud Rate differente\n");
	else
		PDEBUG("io_serial:get_serial_prop->Baud Rate = %s\n", baud_to_string(serial.property.bitrate));
	
	/*	Character size*/
	if( serial.property.bits != C_CSIZE(io_serial.tty) )
		PDEBUG("io_serial:get_serial_prop->Character size differente della seriale\n");
	else
		PDEBUG("io_serial:get_serial_prop->Character size = CS%d\n", (serial.property.bits >> 3)/2+5);
	
	/*	Parity	*/
	if( serial.property.parity != IO_SERIAL_PARITY_NONE ){
		/*	Controllo Dispari	*/
		if( (serial.property.parity == IO_SERIAL_PARITY_ODD) && C_PARODD(io_serial.tty) )
			PDEBUG("io_serial:get_serial_prop->parity: ODD\n");
		/*	Controllo Pari	*/
		else if( (serial.property.parity == IO_SERIAL_PARITY_EVEN) && ~C_PARODD(io_serial.tty) )
			PDEBUG("io_serial:get_serial_prop->parity: EVEN\n");
		else
			PDEBUG("io_serial:get_serial_prop->parity: different!\n");
	}
	else{
		/*	Controllo Nessuno	*/
		if( C_PARENB(io_serial.tty) )
			PDEBUG("io_serial:get_serial_prop->parity: NONE\n");
		else
			PDEBUG("io_serial:get_serial_prop->parity: different!\n");
	}
	
	/*	Controllo Stop Bits	*/
	if( (serial.property.stopbits == 2) && C_CSTOPB(io_serial.tty))
		PDEBUG("io_serial:get_serial_prop->Stop Bits = 2\n");
	else if( (serial.property.stopbits == 1) && ~C_CSTOPB(io_serial.tty))
		PDEBUG("io_serial:get_serial_prop->Stop Bits = 1\n");
	else
		PDEBUG("io_serial:get_serial_prop->Stop Bits differenti!\n");
	
	/*	Controlli obbligatori	*/
	PDEBUG("io_serial:get_serial_prop->ICANON:%1d\n", io_serial.tty->icanon);
	
	/*	Preparazione risposta	*/
	*newprop = serial.property;
	return OK;
}

/*	Setting della porta per comunicare con il programmatore	*/
static int io_smartk_set_property(struct smartk * core, const struct io_property *newprop){
	struct termios newtermios = *io_serial.tty->termios;
	unsigned long modembits;
	mm_segment_t oldfs;
	int ret;
	
	/* 	Setting bitrate per input e output (sotto linux e' lo stesso) 
	*	preso direttamente dalle librerie glibc per sostituire il comando
	*	"cfsetispeed" eseguito dalle lib del towitoko. Per quanto riguarda
	*	"cfsetospeed" si eseguono le stesse ultime 2 operazioni. 
	*/
	/*	Questi cambi vengono effettuati da rs_set_termios chiamato 
	*	automaticamente da n_tty_ioctl
	*/
	
	//newtermios.c_iflag &= ~IBAUD0; //puliamo i 
    newtermios.c_cflag &= ~(CBAUD | CBAUDEX);  //Clean Baud Speed Mask 
    newtermios.c_cflag |= newprop->bitrate;
    serial.property.bitrate = newprop->bitrate;
	
	/*	Setting character size 	*/
  	newtermios.c_cflag |= newprop->bits;
	serial.property.bits = newprop->bits;
	
	/*	Setting parity	*/
	switch (newprop->parity){
    	case IO_SERIAL_PARITY_ODD:
      		newtermios.c_cflag |= PARENB;//attiva parity checking
      		newtermios.c_cflag |= PARODD;
      		break;
    	case IO_SERIAL_PARITY_EVEN:
      		newtermios.c_cflag |= PARENB;
      		newtermios.c_cflag &= ~PARODD;
      		break;
    	case IO_SERIAL_PARITY_NONE:
      		newtermios.c_cflag &= ~PARENB;
    }
	serial.property.parity = newprop->parity;
	
	/* Setting Stop Bits*/
	switch (newprop->stopbits){
    	case 1:
      		newtermios.c_cflag &= (~CSTOPB);
      		break;
    	case 2:
      		newtermios.c_cflag |= CSTOPB;
      		break;
    }
	serial.property.stopbits = newprop->stopbits;
	
	/* Raw (non-canonical) input and output */
	newtermios.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);
  	newtermios.c_oflag &= ~OPOST;
#if 1
  	/* Ignore parity errors!!! Windows driver does so why shouldn't I? */
  	newtermios.c_iflag |= IGNPAR;
#endif
  	/* Enable receiber, hang on close, ignore control line */
  	newtermios.c_cflag |= CREAD | HUPCL | CLOCAL;

  	/* Read 1 byte minimun, no timeout specified */
  	newtermios.c_cc[VMIN] = 1;
  	newtermios.c_cc[VTIME] = 0;

	/* 	Si deve andare in memoria utente per utilizzare le funzioni 
	*	ioclt del driver e di n_tty_ioctl	
	*/
	oldfs = get_fs();
    set_fs(get_ds());
	/*Setting DTR*/
	modembits = TIOCM_DTR;
	
	if (newprop->dtr == IO_SERIAL_HIGH)
		ret = io_serial.tty->driver.ioctl(io_serial.tty, io_serial.file, TIOCMBIS, (unsigned long)&modembits);
	else
		ret = io_serial.tty->driver.ioctl(io_serial.tty, io_serial.file, TIOCMBIC, (unsigned long)&modembits);
	
	if (ret < 0){
		PDEBUG("io_serial:SetProperto	ies->Setting DTR Error:%d!\n", ret);	
		return FAIL;
	}
	serial.property.rts = newprop->rts;
	
	/*Setting RTS*/
	modembits = TIOCM_RTS;
	
	if (newprop->dtr == IO_SERIAL_HIGH)
		ret = io_serial.tty->driver.ioctl(io_serial.tty, io_serial.file, TIOCMBIS, (unsigned long)&modembits);
	else
		ret = io_serial.tty->driver.ioctl(io_serial.tty, io_serial.file, TIOCMBIC, (unsigned long)&modembits);
	
	if (ret < 0){
		PDEBUG("io_serial:SetProperties->Setting RTS Error:%d!\n", ret);	
		return FAIL;
	}
	serial.property.dtr = newprop->dtr;
	
	/*	Setting termios	*/
	ret = n_tty_ioctl(io_serial.tty, io_serial.file, TCSETS, (unsigned long)&newtermios);
	if (ret < 0){	
		PDEBUG("io_serial:SetProperties->Setting Termios Error:%d!\n", ret);	
		return FAIL;
	}

	/*	Torniamo in memoria kernel	*/	
	set_fs(oldfs);
	
	/*	Settiamo la modalita' raw	*/
	io_serial.tty->icanon = (L_ICANON(io_serial.tty) != 0);
  	io_serial.tty->raw = 1;
    io_serial.tty->real_raw = 1;
    PDEBUG("io_serial:SetProperties->Setting RAW!\n");	
  	
/*	Log dei set sulla seriale*/
	PDEBUG("io_serial:set_serial_prop->bitrate=%s bits=CS%1d stopbits=%1d parity=%s dtr=%1d rts=%1d\n",\
			baud_to_string( newprop->bitrate), ((newprop->bits >> 3)/2+5),\
			newprop->stopbits, newprop->parity == IO_SERIAL_PARITY_EVEN ? "EVEN" :\
			newprop->parity == IO_SERIAL_PARITY_ODD ? "ODD": "NONE", newprop->dtr, newprop->rts);
	
	return OK;
}

/*	Scaricamento del modulo	*/
static void io_smartk_cleanup(void)
{
	int ret;

	/* Chiusura della seriale*/
	io_serial.fops->release(io_serial.inode, io_serial.file);
	
	/* Eliminiamo il line discipline*/
	tty_register_ldisc(N_MOUSE, NULL);

	printk(KERN_INFO "io_serial:cleanup->IO Layer Off!\n");
}

/*Definizione init dei moduli*/
module_init(io_smartk_init);
module_exit(io_smartk_cleanup);
